from __future__ import annotations

import base64
import os.path
import time
from datetime import datetime, timezone
from email.message import EmailMessage
from email.utils import parseaddr
from functools import lru_cache
from typing import Dict, List
from urllib.error import HTTPError

import requests
import tenacity
from google.auth.exceptions import RefreshError
from google.oauth2.credentials import Credentials
from googleapiclient.discovery import build
from loguru import logger
from pydantic import EmailStr, Field

from plurally.models.action.email_smtp import SendEmailInputSchema
from plurally.models.misc import Table
from plurally.models.node import Node
from plurally.models.source.email_imap import (
    EmailSchema,
    EmailSourceInputSchema,
    EmailSourceOutputSchema,
)


def parse_email(gmail_response: dict) -> EmailSchema:
    payload = gmail_response.get("payload", {})
    headers = payload.get("headers", [])

    # Helper function to extract a header value
    def get_header(name):
        for header in headers:
            if header["name"].lower() == name.lower():
                return header["value"]
        return None

    # Extract the sender's name and email from the 'From' header
    from_header = get_header("From")
    if from_header:
        # Parsing the 'From' header (usually in the format 'Name <email@example.com>')
        sender_name, sender_email = parseaddr(from_header)
    else:
        sender_name, sender_email = None, None

    # Extract subject
    subject = get_header("Subject") or ""

    # Extract the body content (handling text/plain)
    content = ""
    if "body" in payload and payload["body"].get("data"):
        content = payload["body"]["data"]
    else:
        # If the email has parts (e.g., multipart/alternative), search for the plain text part
        parts = payload.get("parts", [])
        for part in parts:
            if part["mimeType"] == "text/plain" and "data" in part["body"]:
                content = part["body"]["data"]
                break

    # Extract datetime received from 'internalDate'
    internal_date = gmail_response.get("internalDate")
    datetime_received = (
        datetime.fromtimestamp(int(internal_date) / 1000)
        if internal_date
        else datetime.now(timezone.utc)
    )
    datetime_received = datetime_received.replace(tzinfo=None)

    # Build and return the Pydantic model
    return EmailSchema(
        sender_name=sender_name or "Unknown",
        sender_email=sender_email or "unknown@example.com",
        subject=subject,
        content=content,
        datetime_received=datetime_received,
    )


COMMON_DESC = """```info
This block requires you to connect your GMail account to Plurally.
```"""


class Google(Node):
    SCOPES: List[str] = ["openid", "email"]
    SERVICE: str = None
    VERSION: str = None

    class InitSchema(Node.InitSchema): ...

    class InputSchema(Node.InputSchema): ...

    def __init__(self, init_inputs: InitSchema, outputs=None):
        assert self.SCOPES is not None, "SCOPES must be defined in the subclass"
        assert self.SERVICE is not None, "SERVICE must be defined in the subclass"
        assert self.VERSION is not None, "VERSION must be defined in the subclass"

        self._token = None
        self._token_expiry = None
        self._service = None
        self._tokens_params: Dict[str, str] = {}

        super().__init__(init_inputs, outputs)

    def token(self):
        now = datetime.now(tz=timezone.utc).replace(tzinfo=None)
        if self._token is None or self._token_expiry < now:
            self.reset()
            self._token, self._token_expiry = self._get_access_token()
        return self._token

    def _get_access_token(self):
        token_url = os.environ.get("PLURALLY_TOKEN_URL")
        assert token_url, "PLURALLY_TOKEN_URL must be set in the environment"

        api_key = os.environ.get("PLURALLY_API_KEY")
        assert api_key, "PLURALLY_API_KEY must be set in the environment"

        headers = {
            "Authorization": f"Bearer {api_key}",
        }
        res = requests.get(
            token_url,
            headers=headers,
            params={"scopes": " ".join(self.SCOPES), **self._tokens_params},
        )
        res.raise_for_status()

        data = res.json()
        token_expiry = datetime.fromisoformat(data["expires_at"])
        return data["access_token"], token_expiry

    @property
    def service(self):
        if self._service is None:
            creds = Credentials(token=self.token())
            self._service = build(self.SERVICE, self.VERSION, credentials=creds)
        return self._service

    def reset(self):
        self._token = None
        self._token_expiry = None
        self._service = None

    @tenacity.retry(
        stop=tenacity.stop_after_attempt(2),
        retry=tenacity.retry_if_exception_type(RefreshError),
    )
    def forward(self, input_schema: InputSchema):
        try:
            self._forward(input_schema)
        except RefreshError:
            # resetting will request a new token
            self.reset()
            raise


class GmailRead(Google):
    ICON = "gmail"
    SCOPES = list(
        set(Google.SCOPES + ["https://www.googleapis.com/auth/gmail.readonly"])
    )
    SERVICE = "gmail"
    VERSION = "v1"

    SensitiveFields = ("email",)

    class InitSchema(Google.InitSchema):

        email: EmailStr = Field(
            title="Email",
            description="The email address to read emails from.",
            examples=["yourname@gmail.com"],
        )

    class InputSchema(Node.InputSchema): ...

    def __init__(self, init_inputs: Google.InitSchema):
        self.email = init_inputs.email
        self._tokens_params = {"email": self.email}
        super().__init__(init_inputs)

    def serialize(self):
        return super().serialize() | {"email": self.email}

    def get_scopes_and_meta_data(self):
        return [(self.SCOPES, {"email": self.email})]


class GMailNewEmail(GmailRead):
    ICON = "gmail"
    IS_TRIGGER = True
    OutputSchema = EmailSchema

    class InitSchema(GmailRead.InitSchema):
        __doc__ = f"""Trigger the flow when a new email is received in a mailbox from a GMail address.
```warning
If you wish to process multiple emails or use them as source data, use the **kls::GmailSource** node instead.
```
{COMMON_DESC}"""

    DESC = InitSchema.__doc__

    def __init__(self, init_inputs: InitSchema):
        self.check_after = datetime.now(tz=timezone.utc).replace(tzinfo=None)
        super().__init__(init_inputs)

    def _forward(self, _):
        # Format self.check_after to YYYY/MM/DD for the Gmail query
        formatted_check_after = self.check_after.strftime("%Y/%m/%d")

        # Fetch unread messages received after `self.check_after`
        query = f"is:unread after:{formatted_check_after}"
        messages = self.service.users().messages().list(userId="me", q=query).execute()

        # If no unread messages are found, return an empty result
        if "messages" not in messages:
            self.outputs = None
            return

        email_ids = [msg["id"] for msg in messages["messages"]]
        emails = []

        for email_id in email_ids:
            email = (
                self.service.users().messages().get(userId="me", id=email_id).execute()
            )

            # Add email along with its received date (internalDate)
            emails.append(email)

        for email in emails:
            email_received_timestamp = (
                int(email["internalDate"]) / 1000
            )  # Convert from ms to seconds
            email_received_datetime = datetime.fromtimestamp(
                email_received_timestamp
            ).replace(tzinfo=None)

            if email_received_datetime > self.check_after:
                self.outputs = parse_email(email).model_dump()
                return

        self.outputs = None


class GMailSource(GmailRead):

    class InitSchema(GmailRead.InitSchema):
        __doc__ = f"""Read emails in a mailbox from a GMail address.

```warning
If you want to trigger the flow when a new email is received, use the **kls::GmailNewEmail** node instead.
```
{COMMON_DESC}"""

        limit: int = Field(
            default=200,
            ge=1,
            le=200,
            title="Max Emails",
            examples=[200],
            description="The maximum number of emails to process.",
        )

    DESC = InitSchema.__doc__

    InputSchema = EmailSourceInputSchema
    OutputSchema = EmailSourceOutputSchema

    def __init__(self, init_inputs: "InitSchema"):
        self.limit = init_inputs.limit
        super().__init__(init_inputs)

    def serialize(self):
        return super().serialize() | {"limit": self.limit}

    @lru_cache(maxsize=200)
    def _fetch_email(self, email_id: str):
        email = self.service.users().messages().get(userId="me", id=email_id).execute()
        return email

    def _build_query(self, range_start: datetime, range_end: datetime):
        formatted_check_after = range_start.strftime("%Y/%m/%d")
        query = f"after:{formatted_check_after}"

        formatted_check_before = range_end.strftime("%Y/%m/%d")

        if query is None:
            query = f"before:{formatted_check_before}"
        else:
            query = f"{query} before:{formatted_check_before}"
        return query

    @lru_cache(2)
    def _fetch_email_ids(self, query: str, ttl_hash: int):
        messages = (
            self.service.users()
            .messages()
            .list(userId="me", maxResults=self.limit, q=query)
            .execute()
        )

        messages = messages.get("messages", [])
        email_ids = [msg["id"] for msg in messages[: self.limit]]
        return email_ids

    def _forward(self, node_input: "InputSchema"):
        query = self._build_query(node_input.range_start, node_input.range_end)
        cache_secs = 60
        email_ids = self._fetch_email_ids(
            query, ttl_hash=time.time() // cache_secs
        )  # TTL of 1 minute
        emails = []
        range_start_timestamp = (
            node_input.range_start.timestamp() // cache_secs * cache_secs
        )
        range_end_timestamp = (
            node_input.range_end.timestamp() // cache_secs * cache_secs
        )
        for email_id in email_ids:
            email = self._fetch_email(email_id)

            email_received_timestamp = int(email["internalDate"]) / 1000
            if email_received_timestamp < range_start_timestamp:
                return None
            if email_received_timestamp > range_end_timestamp:
                return None

            emails.append(parse_email(email).model_dump())
            if len(emails) == self.limit:
                break

        logger.debug(f"Processed {len(emails)} emails")
        self.outputs = {"emails": Table(data=emails)}


class GMailSend(Google):
    ICON = "gmail"
    SCOPES = list(set(Google.SCOPES + ["https://www.googleapis.com/auth/gmail.send"]))
    SERVICE = "gmail"
    VERSION = "v1"
    SensitiveFields = ("email",)

    class InitSchema(Google.InitSchema):
        __doc__ = f"""Send email using Gmail.

{COMMON_DESC}"""

        email: EmailStr = Field(
            title="Email",
            description="The email address to read emails from.",
            examples=["yourname@gmail.com"],
        )
        fullname: str = Field(
            "",
            title="Full Name",
            description="The full name of the sender.",
            examples=["John Doe"],
        )

    DESC = InitSchema.__doc__

    InputSchema = SendEmailInputSchema

    class OutputSchema(Node.OutputSchema): ...

    def __init__(self, init_inputs: InitSchema):
        self.email = init_inputs.email
        self.fullname = init_inputs.fullname
        self._tokens_params = {"email": self.email}
        super().__init__(init_inputs)

    @property
    def from_email(self):
        return f"{self.fullname} <{self.email}>" if self.fullname else self.email

    def _forward(self, node_input: InputSchema):
        message = EmailMessage()
        message.set_content(node_input.body)

        message["To"] = node_input.email_address
        message["From"] = self.from_email
        message["Subject"] = node_input.subject
        encoded_message = base64.urlsafe_b64encode(message.as_bytes()).decode()

        create_message = {"raw": encoded_message}
        send_message = (
            self.service.users()
            .messages()
            .send(userId="me", body=create_message)
            .execute()
        )
        logger.debug(f"Sent email with ID: {send_message['id']}")

    def serialize(self):
        return super().serialize() | {
            "email": self.email,
            "fullname": self.fullname,
        }

    def get_scopes_and_meta_data(self):
        return [(self.SCOPES, {"email": self.email})]


class SheetsRead(Google):
    ICON = "gsheets"
    SCOPES = ["https://www.googleapis.com/auth/spreadsheets.readonly"]
    SERVICE = "sheets"
    VERSION = "v4"

    SensitiveFields = ("sheet_id", "range_name")

    class InitSchema(Google.InitSchema):
        """Read data from a Google Sheet."""

        sheet_id: str = Field(
            title="Sheet ID",
            description="The ID of the Google Sheet.",
            examples=["1CM29gwKIzeXsAppeNwrc8lbYaVcmUclprLuLYuHog4k"],
        )
        range_name: str = Field(
            default="Sheet1",
            title="Range Name",
            description="The range of the sheet to read.",
            examples=["Sheet1", "Sheet1!A1:B2"],
        )

    class InputSchema(Node.InputSchema): ...

    class OutputSchema(Node.OutputSchema):
        values: Table = Field(
            title="Values",
            description="The values in the spreadsheet.",
        )

    DESC = InitSchema.__doc__

    def __init__(self, init_inputs: InitSchema, outputs=None):
        self.sheet_id = init_inputs.sheet_id
        self.range_name = init_inputs.range_name
        super().__init__(init_inputs, outputs)

    def _read_values(self):
        sheet = self.service.spreadsheets()
        try:
            result = (
                sheet.values()
                .get(spreadsheetId=self.sheet_id, range=self.range_name)
                .execute()
            )
        except HTTPError as e:
            if e.response.status_code == 403:
                raise ValueError("You do not have permission to read this sheet")
            elif e.response.status_code == 404:
                raise ValueError(str(e))
            raise
        return result.get("values", [])

    def _forward(self, _: InputSchema):
        values = self._read_values()
        # turn into list of dicts
        if values:
            values = [dict(zip(values[0], row)) for row in values[1:]]

        self.outputs = {"values": Table(data=values)}

    def serialize(self):
        return super().serialize() | {
            "sheet_id": self.sheet_id,
            "range_name": self.range_name,
        }


class SheetsWrite(Google):
    ICON = "gsheets"
    SCOPES = ["https://www.googleapis.com/auth/spreadsheets"]
    SERVICE = "sheets"
    VERSION = "v4"
    SensitiveFields = ("sheet_id", "range_name")

    class InitSchema(Google.InitSchema):
        """Read data from a Google Sheet."""

        sheet_id: str = Field(
            title="Sheet ID",
            description="The ID of the Google Sheet.",
            examples=["1CM29gwKIzeXsAppeNwrc8lbYaVcmUclprLuLYuHog4k"],
        )
        append: bool = Field(
            True,
            title="Append",
            description="Whether to append to the sheet or overwrite.",
        )
        range_name: str = Field(
            default="Sheet1",
            title="Range Name",
            description="The range of the sheet to write.",
            examples=["Sheet1", "Sheet1!A1:B2"],
        )

    class InputSchema(Node.InputSchema):
        values: Table

    class OutputSchema(Node.OutputSchema): ...

    DESC = InitSchema.__doc__

    def __init__(self, init_inputs: InitSchema, outputs=None):
        self.sheet_id = init_inputs.sheet_id
        self.append = init_inputs.append
        self.range_name = init_inputs.range_name
        self.value_input_option = "RAW"

        super().__init__(init_inputs, outputs)

    def _write_values(self, data: List[Dict[str, str]]):
        if not data:
            logger.debug("No data to write")
            return

        sheet = self.service.spreadsheets()
        headers = list(data[0].keys())

        # Get the first row of the sheet to check if headers are present
        result = (
            sheet.values()
            .get(
                spreadsheetId=self.sheet_id,
                range=self.range_name,
                majorDimension="ROWS",
            )
            .execute()
        )
        current_values = result.get("values", [])

        rows = [
            list(data[0].keys()),
            *[[row[col] for col in data[0].keys()] for row in data],
        ]
        if current_values and current_values[0] == headers:
            logger.debug("Headers already present, skipping header row")
            body = {"values": rows[1:]}
        else:
            logger.debug("Headers not present, writing headers and data")
            body = {"values": rows}

        try:
            if self.append:
                sheet.values().append(
                    spreadsheetId=self.sheet_id,
                    range=self.range_name,
                    valueInputOption=self.value_input_option,
                    body=body,
                ).execute()
            else:
                sheet.values().update(
                    spreadsheetId=self.sheet_id,
                    range=self.range_name,
                    valueInputOption=self.value_input_option,
                    body=body,
                ).execute()
        except HTTPError as e:
            if e.response.status_code == 403:
                raise ValueError("You do not have permission to write to this sheet")
            raise

    def _forward(self, node_input: InputSchema):
        self._write_values(node_input.values.data)

    def serialize(self):
        return super().serialize() | {
            "sheet_id": self.sheet_id,
            "range_name": self.range_name,
            "append": self.append,
        }
